"""
Tasseled cap index Percentiles
"""
from functools import partial
from typing import Sequence, Tuple, Iterable, Dict
import xarray as xr
import numpy as np
import logging
from odc.algo import keep_good_only
from odc.algo._percentile import xr_quantile_bands
from odc.algo._masking import (
    _xr_fuse,
    _fuse_mean_np,
    enum_to_bool,
    mask_cleanup,
)
from ._registry import StatsPluginInterface, register

_log = logging.getLogger(__name__)
NODATA = -9999  # output NODATA


class StatsTCWPC(StatsPluginInterface):
    NAME = "ga_tc_percentiles"
    SHORT_NAME = NAME
    VERSION = "1.0.1"
    PRODUCT_FAMILY = "tci"

    def __init__(
        self,
        coefficients: Dict[str, Dict[str, float]] = None,
        input_bands: Sequence[str] = None,
        output_bands: Sequence[str] = None,
        cloud_filters: Dict[str, Iterable[Tuple[str, int]]] = None,
        **kwargs,
    ):
        self.cloud_filters = cloud_filters
        if coefficients is None:
            self.coefficients = dict(
                [
                    (
                        "wet",
                        dict(
                            [
                                ("blue", 0.0315),
                                ("green", 0.2021),
                                ("red", 0.3102),
                                ("nir", 0.1594),
                                ("swir1", -0.6806),
                                ("swir2", -0.6109),
                            ]
                        ),
                    ),
                    (
                        "bright",
                        dict(
                            [
                                ("blue", 0.2043),
                                ("green", 0.4158),
                                ("red", 0.5524),
                                ("nir", 0.5741),
                                ("swir1", 0.3124),
                                ("swir2", 0.2303),
                            ]
                        ),
                    ),
                    (
                        "green",
                        dict(
                            [
                                ("blue", -0.1603),
                                ("green", -0.2819),
                                ("red", -0.4934),
                                ("nir", 0.7940),
                                ("swir1", -0.0002),
                                ("swir2", -0.1446),
                            ]
                        ),
                    ),
                ]
            )
        self.input_bands = (
            [
                "blue",
                "green",
                "red",
                "nir",
                "swir1",
                "swir2",
                "fmask",
                "nbart_contiguity",
            ]
            if input_bands is None
            else input_bands
        )

        super().__init__(input_bands=self.input_bands, **kwargs)
        self.output_bands = (
            ["wet", "bright", "green"] if output_bands is None else output_bands
        )

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurments = []
        for band in self.output_bands:
            _measurments += [f"{band}_pc_10", f"{band}_pc_50", f"{band}_pc_90"]
        return _measurments

    def native_transform(self, xx):
        """
        Loads data in its native projection.
        """
        mask = xx["fmask"]
        nodata = enum_to_bool(mask, ("nodata",))
        non_contiguent = xx["nbart_contiguity"] == 0
        bad = nodata | non_contiguent
        # Now exclude cloud and cloud shadow (including buffered pixels)
        if self.cloud_filters is not None:
            for cloud_class, c_filter in self.cloud_filters.items():
                cloud_mask = enum_to_bool(mask, (cloud_class,))
                cloud_mask_buffered = mask_cleanup(cloud_mask, mask_filters=c_filter)
                bad = cloud_mask_buffered | bad
        else:
            _log.info("There is no cloud/shadow buffering.")
            cloud_shadow_mask = enum_to_bool(mask, ("cloud", "shadow"))
            bad = bad | cloud_shadow_mask

        for band in xx.data_vars.keys():
            bad = bad | (xx[band] == -999)

        yy = xx.drop_vars(self.input_bands)
        for m in self.output_bands:
            yy[m] = sum(
                coeff * xx[band] for band, coeff in self.coefficients[m].items()
            ).astype(np.int16)
            yy[m].attrs = xx.blue.attrs
            yy[m].attrs["nodata"] = NODATA

        yy = keep_good_only(yy, ~bad, nodata=NODATA)
        return yy

    def fuser(self, xx):
        xx = _xr_fuse(xx, partial(_fuse_mean_np, nodata=NODATA), "")

        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        yy = xr_quantile_bands(xx, [0.1, 0.5, 0.9], nodata=NODATA)
        return yy


register("tcw-percentiles", StatsTCWPC)



"""
Mangroves canopy cover classes
"""
from typing import Tuple

import numpy as np
import xarray as xr
import dask
import os
from odc.algo import keep_good_only, erase_bad
import fiona
from rasterio import features

from ._registry import StatsPluginInterface, register

NODATA = 255


class Mangroves(StatsPluginInterface):
    NAME = "mangroves"
    SHORT_NAME = NAME
    VERSION = "0.0.2"
    PRODUCT_FAMILY = "mangroves"

    def __init__(
        self,
        pv_thresholds=None,
        tcw_threshold=-1850,
        **kwargs,
    ):
        if pv_thresholds is None:
            pv_thresholds = [14, 38, 60]
        self.mangroves_extent = kwargs.pop("mangroves_extent", None)
        self.pv_thresholds = pv_thresholds
        self.tcw_threshold = tcw_threshold
        super().__init__(input_bands=["pv_pc_10", "qa", "wet_pc_10"], **kwargs)

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurements = ["canopy_cover_class"]
        return _measurements

    def rasterize_mangroves_extent(self, shape_file, transform, dst_shape):
        with fiona.open(shape_file) as source_ds:
            geoms = [s["geometry"] for s in source_ds]

        mangrove_extent = features.rasterize(
            geoms,
            transform=transform,
            out_shape=dst_shape[1:],
            all_touched=False,
            fill=0,
            default_value=1,
            dtype="uint8",
        )

        return dask.array.from_array(mangrove_extent.reshape(dst_shape), name=False)

    def fuser(self, xx):
        """
        no fuse required for mangroves since group by none
        return loaded data
        """
        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        """
        mangroves computation here
        it is not a 'reduce' though
        """
        if self.mangroves_extent:
            if not os.path.exists(self.mangroves_extent):
                raise FileNotFoundError(f"{self.mangroves_extent} not found")
            extent_mask = self.rasterize_mangroves_extent(
                self.mangroves_extent, xx.geobox.transform, xx.pv_pc_10.shape
            )
        else:
            extent_mask = dask.array.ones(xx.pv_pc_10.shape)
        good_data = extent_mask == 1
        good_data &= xx.wet_pc_10 > self.tcw_threshold
        good_data &= (xx.pv_pc_10 > self.pv_thresholds[0]) & (xx.qa == 2) | (xx.qa == 1)

        notsure_mask = xx.qa == 1

        cover_type = xx.pv_pc_10.copy(True)
        cover_type.data = dask.array.zeros_like(cover_type.data)
        for s_t in self.pv_thresholds:
            cover_type.data += (xx.pv_pc_10.data > s_t).astype(np.uint8)

        cover_type = erase_bad(cover_type, notsure_mask, nodata=0)
        cover_type = keep_good_only(cover_type, good_data, nodata=NODATA)
        cover_type.attrs["nodata"] = NODATA

        cover_type = cover_type.to_dataset(name="canopy_cover_class")
        # don't want the dimension spec from input but keep the info in case
        if "spec" in cover_type.dims:
            cover_type = cover_type.squeeze(dim=["spec"])
        return cover_type


register("mangroves", Mangroves)



import json
import sys
from typing import List, Optional

import click
import fsspec
import toolz
from datacube.utils.geometry import Geometry
from odc.aws.queue import get_queue, publish_messages
from odc.dscache.tools.tiling import GRIDS
from odc.stats.model import TileIdx_txy
from odc.stats.tasks import TaskReader, render_sqs

from ._cli_common import main, parse_all_tasks


def do_dry_run(tasks):
    for period, ix, iy in tasks:
        print(f"{period}/{ix:+04d}/{iy:+04d}")


def get_geometry(geojson_file: str) -> Geometry:
    with fsspec.open(geojson_file) as f:
        data = json.load(f)

    return Geometry(
        data["features"][0]["geometry"], crs=data["crs"]["properties"]["name"]
    )


def filter_tasks(tasks: List[TileIdx_txy], geometry: Geometry, grid_name: str):
    for task in tasks:
        task_geometry = GRIDS[grid_name].tile_geobox((task[1], task[2])).extent
        if task_geometry.intersects(geometry):
            yield task


def publish_tasks(
    db: str, task_filter: str, geojson_filter: Optional[str], dryrun: bool, queue: str
):
    reader = TaskReader(db)
    if len(task_filter) == 0:
        tasks = reader.all_tiles
        print(f"Found {len(tasks):,d} tasks in the file")
    else:
        try:
            tasks = parse_all_tasks(task_filter, reader.all_tiles)
            print(
                f"Found {len(tasks):,d} tasks in the file after filtering with {task_filter}"
            )
        except ValueError as e:
            print(str(e), file=sys.stderr)
            sys.exit(1)

    if geojson_filter is not None:
        geometry = get_geometry(geojson_filter)
        tasks = list(filter_tasks(tasks, geometry, reader.grid))
        print(
            f"Found {len(tasks):,d} tasks in the file after filtering with GeoJSON feature(s)"
        )

    if dryrun:
        do_dry_run(tasks)
        sys.exit(0)

    queue = get_queue(queue)

    # We assume the db files are always be the S3 uri. If they are not, there is no need to use SQS queue to process.
    messages = (
        dict(Id=str(idx), MessageBody=json.dumps(render_sqs(tidx, db)))
        for idx, tidx in enumerate(tasks)
    )

    for bunch in toolz.partition_all(10, messages):
        publish_messages(queue, bunch)


@main.command("publish-tasks")
@click.argument("db", type=str)
@click.argument("queue", type=str)
@click.option(
    "--dryrun", is_flag=True, help="Do not publish just print what would be submitted"
)
@click.option("--geojson-filter", help="GeoJSON file to use as a geometry filter")
@click.argument("task_filter", type=str, nargs=-1)
def publish_to_queue(db, queue, dryrun, geojson_filter, task_filter):
    """
    Publish tasks to SQS.

    Task filter can be one of the 3 things

    \b
    1. Comma-separated triplet: period,x,y or 'x[+-]<int>/y[+-]<int>/period
       2019--P1Y,+003,-004
       2019--P1Y/3/-4          `/` is also accepted
       x+003/y-004/2019--P1Y   is accepted as well
    2. A zero based index
    3. A slice following python convention <start>:<stop>[:<step]
        ::10 -- every tenth task: 0,10,20,..
       1::10 -- every tenth but skip first one 1, 11, 21 ..
        :100 -- first 100 tasks

    If no tasks are supplied all tasks will be published the queue.
    """

    publish_tasks(db, task_filter, geojson_filter, dryrun, queue)



import random
from typing import Optional, Tuple, Union, Callable, Any, Dict, List, Iterable, Iterator
from types import SimpleNamespace
from collections import namedtuple
from datetime import datetime
from itertools import islice, chain, groupby
import pickle
import json
import os
from urllib.parse import urlparse
import logging
import ciso8601
import re

from odc.dscache import DatasetCache
from datacube import Datacube
from datacube.model import Dataset, GridSpec, DatasetType
from datacube.utils.geometry import Geometry
from datacube.utils.documents import transform_object_tree
from datacube.utils.dates import normalise_dt

from odc.dscache.tools import bin_dataset_stream, ordered_dss
from odc.dscache.tools.tiling import parse_gridspec_with_name
from odc.dscache.tools.profiling import ds_stream_test_func
from ._text import split_and_check

from odc.aws import s3_download, s3_url_parse

from .model import DateTimeRange, Task, OutputProduct, TileIdx, TileIdx_txy, TileIdx_xy
from ._gjson import gs_bounds, compute_grid_info, gjson_from_tasks
from .utils import (
    bin_annual,
    bin_full_history,
    bin_generic,
    bin_seasonal,
    bin_rolling_seasonal,
    fuse_ds,
    fuse_products,
)

TilesRange2d = Tuple[Tuple[int, int], Tuple[int, int]]
CompressedDataset = namedtuple("CompressedDataset", ["id", "time"])


def _xy(tidx: TileIdx) -> TileIdx_xy:
    return tidx[-2:]


def compress_ds(ds: Dataset) -> CompressedDataset:
    dt = normalise_dt(ds.center_time)
    return CompressedDataset(ds.id, dt)


def is_tile_in(tidx: Tuple[int, int], tiles: TilesRange2d) -> bool:
    (x0, x1), (y0, y1) = tiles
    x, y = tidx
    return (x0 <= x < x1) and (y0 <= y < y1)


def out_path(suffix: str, base: str) -> str:
    if base.endswith(".db"):
        base = base[:-3]
    return base + suffix


def sanitize_query(query):
    def sanitize(v):
        if isinstance(v, Geometry):
            return v.json
        if isinstance(v, datetime):
            return v.isoformat()
        return v

    return transform_object_tree(sanitize, query)


def render_task(tidx: TileIdx_txy) -> str:
    period, xi, yi = tidx
    return f"{period}/{xi:02d}/{yi:02d}"


def parse_task(s: str) -> TileIdx_txy:
    sep = "/" if "/" in s else ","
    t, x, y = split_and_check(s, sep, 3)
    if t.startswith("x"):
        t, x, y = y, t, x
    return (t, int(x.lstrip("x")), int(y.lstrip("y")))


def render_sqs(tidx: TileIdx_txy, filedb: str) -> Dict[str, str]:
    """
    Add extra layer to render task. Convert it to JSON for SQS message body.
    """
    period, xi, yi = tidx
    return {"filedb": filedb, "tile_idx": f"{period}/{xi:02d}/{yi:02d}"}


def parse_sqs(s: str) -> Tuple[TileIdx_txy, str]:
    """
    Add extra layer to parse task. Convert it from JSON for SQS message body.
    """

    message_body = json.loads(s)
    filedb = message_body.get("filedb", None)

    tile_info = message_body.get("tile_idx", None)

    sep = "/" if "/" in s else ","
    t, x, y = split_and_check(tile_info, sep, 3)
    if t.startswith("x"):
        t, x, y = y, t, x
    return ((t, int(x.lstrip("x")), int(y.lstrip("y"))), filedb)


class SaveTasks:
    def __init__(
        self,
        output: str,
        grid: str,
        frequency: str = "annual",
        overwrite: bool = False,
        complevel: int = 6,
    ):
        if DatasetCache.exists(output) and overwrite is False:
            raise ValueError(f"File database already exists: {output}")

        grid, gridspec = parse_gridspec_with_name(grid)

        self._output = output
        self._overwrite = overwrite
        self._complevel = complevel
        self._grid = grid
        self._gridspec = gridspec
        self._frequency = frequency

    def out_path(self, suffix: str) -> str:
        return out_path(suffix, self._output)

    @classmethod
    def ds_align(
        cls,
        dss: Iterable,
        products: List[DatasetType],
        group_size: int,
        fuse_dss: bool = True,
    ):
        def match_dss(groups, group_size):
            for _, ds_group in groups:
                ds_group = tuple(ds_group)
                if len(ds_group) >= group_size:
                    yield ds_group

        grouped_dss = groupby(
            dss,
            key=lambda ds: (ds.center_time, ds.metadata.region_code)
            if hasattr(ds.metadata, "region_code")
            else (ds.center_time,),
        )
        grouped_dss = match_dss(grouped_dss, group_size)

        if fuse_dss:
            fused_product = fuse_products(*products)

            def map_fuse_func(x):
                return fuse_ds(*x, product=fused_product)

            dss = map(map_fuse_func, grouped_dss)
            return dss
        else:
            return grouped_dss

    @classmethod
    def _find_dss(
        cls,
        dc: Datacube,
        products: str,
        query: Dict[str, Any],
        dataset_filter=None,
        predicate=None,
        fuse_dss: bool = True,
    ):
        """
        query and filter the datasets with a string composed by products name
        A string joined by `-` implies union of all datasets
        A string joined by `+` implies intersect (filtered and then groupby) against time
        return a generator of datasets
        """
        if dataset_filter is None:
            dataset_filter = {}

        # query by a list of products is not a "officially" supported feature
        # but it is embedded in the code everywhere
        # mark it for ref

        def sanitize_products_str(products_str):
            """
            split a string composed by product names into a list of product names
            e.g., ga_ls8-ga_ls7 -> [ga_ls8, ga_ls7]
            rules:
            1. Any separator (`+/-`) at the start or the end is disregarded
            2. Multiple same separators between the product names are treated as one
            3. Multiple different separators between the product names is respected by left-right order
            e.g., ga_ls8+-ga_ls7 -> separator is `+` as `+` proceeds `-` from left to right
            """
            product_list = re.split(r"[\+-]{1,}", products_str)
            product_list = list(filter(None, product_list))
            group_size = len(re.findall(r"\w{1,}\+[-\+]{0,}\w{1,}", products_str))
            return product_list, group_size

        product_list, group_size = sanitize_products_str(products)

        query.update(dict(product=product_list, **dataset_filter))
        dss = ordered_dss(
            dc,
            freq="y",
            key=lambda ds: (ds.center_time, ds.metadata.region_code)
            if hasattr(ds.metadata, "region_code")
            else (ds.center_time,),
            **query,
        )

        if group_size > 0:
            products = [
                dc.index.products.get_by_name(product) for product in product_list
            ]
            dss = cls.ds_align(dss, products, group_size + 1, fuse_dss)

        if predicate is not None:
            dss = filter(predicate, dss)

        return dss

    def get_dss_by_grid(
        self,
        dc: Datacube,
        products: str,
        msg: Callable[[str], Any],
        dataset_filter=None,
        predicate=None,
        temporal_range: Optional[DateTimeRange] = None,
        tiles: Optional[TilesRange2d] = None,
    ):
        """
        This returns a tuple containing:
        - a generator of datasets
        - the number of datasets in the generator
        - a config dictionary containing the product, temporal range, tiles, and the datacube query used
        """

        cfg: Dict[str, Any] = dict(
            grid=self._grid,
            freq=self._frequency,
        )

        query = {}

        if tiles is not None:
            (x0, x1), (y0, y1) = tiles
            msg(f"Limit search to tiles: x:[{x0}, {x1}) y:[{y0}, {y1})")
            cfg["tiles"] = tiles
            query["geopolygon"] = gs_bounds(self._gridspec, tiles)

        if temporal_range is not None:
            query.update(
                temporal_range.dc_query(pad=0.6)
            )  # pad a bit more than half a day on each side
            cfg["temporal_range"] = temporal_range.short

        msg("Connecting to the database, streaming datasets")
        dss = self._find_dss(dc, products, query, dataset_filter, predicate)
        cfg["query"] = sanitize_query(query)

        return dss, cfg

    def save(
        self,
        dc: Datacube,
        products: str,
        dataset_filter=None,
        temporal_range: Union[str, DateTimeRange, None] = None,
        tiles: Optional[TilesRange2d] = None,
        predicate: Optional[Callable[[Dataset], bool]] = None,
        msg: Optional[Callable[[str], Any]] = None,
        debug: bool = False,
    ) -> bool:
        """
        :param products: Product name to consume
        :param dataset_filter: Optionally apply search filter on Datasets
        :param temporal_range: Optionally  limit query in time
        :param tiles: Optionally limit query to a range of tiles
        :param predicate: If supplied filter Datasets as they come in with custom filter, Dataset->Bool
        :param msg: Observe messages if needed via callback
        :param debug: Dump some intermediate state to files for debugging
        """
        # pylint:disable=too-many-locals,too-many-branches,too-many-statements

        if DatasetCache.exists(self._output) and self._overwrite is False:
            raise ValueError(f"File database already exists: {self._output}")

        dt_range = SimpleNamespace(start=None, end=None)

        def _update_start_end(x, out):
            if out.start is None:
                out.start = x
                out.end = x
            else:
                out.start = min(out.start, x)
                out.end = max(out.end, x)

        def persist(ds: Dataset) -> CompressedDataset:
            _ds = compress_ds(ds)
            _update_start_end(_ds.time, dt_range)
            return _ds

        def msg_default(msg):
            pass

        if msg is None:
            msg = msg_default

        if isinstance(temporal_range, str):
            temporal_range = DateTimeRange(temporal_range)

        dss, cfg = self.get_dss_by_grid(
            dc, products, msg, dataset_filter, predicate, temporal_range, tiles
        )

        dss_slice = list(islice(dss, 0, 100))
        if len(dss_slice) == 0:
            msg("found no datasets")
            return True

        if len(dss_slice) >= 100:
            msg("Training compression dictionary")
            samples = dss_slice.copy()
            random.shuffle(samples)
            zdict = DatasetCache.train_dictionary(samples, 8 * 1024)
            msg(".. done")
        else:
            zdict = None

        dss = chain(dss_slice, dss)
        cache = DatasetCache.create(
            self._output,
            zdict=zdict,
            complevel=self._complevel,
            truncate=self._overwrite,
        )
        cache.add_grid(self._gridspec, self._grid)
        cache.append_info_dict("stats/", dict(config=cfg))

        cells: Dict[Tuple[int, int], Any] = {}
        dss = cache.tee(dss)
        dss = bin_dataset_stream(self._gridspec, dss, cells, persist=persist)
        rr = ds_stream_test_func(dss)
        msg(rr.text)

        if tiles is not None:
            # prune out tiles that were not requested
            cells = {
                tidx: cell for tidx, cell in cells.items() if is_tile_in(tidx, tiles)
            }

        if temporal_range is not None:
            # Prune Datasets outside of temporal range (after correcting for UTC offset)
            for cell in cells.values():
                utc_offset = cell.utc_offset
                cell.dss = [
                    ds for ds in cell.dss if ds.time + utc_offset in temporal_range
                ]

        n_tiles = len(cells)
        msg(f"Total of {n_tiles:,d} spatial tiles")

        if self._frequency == "all":
            if temporal_range is None:
                tasks = bin_full_history(cells, start=dt_range.start, end=dt_range.end)
            else:
                tasks = bin_generic(cells, [temporal_range])
        elif self._frequency == "semiannual":
            tasks = bin_seasonal(cells, months=6, anchor=1)
        elif self._frequency == "seasonal":
            tasks = bin_seasonal(cells, months=3, anchor=12)
        elif self._frequency == "quartely":
            tasks = bin_seasonal(cells, months=3, anchor=1)
        elif self._frequency == "3month-seasons":
            anchor = int(temporal_range.start.strftime("%m"))
            tasks = bin_seasonal(cells, months=3, anchor=anchor)
        elif self._frequency == "rolling-3months":
            tasks = bin_rolling_seasonal(
                cells, temporal_range=temporal_range, months=3, interval=1
            )
        elif self._frequency == "nov-mar":
            tasks = bin_seasonal(cells, months=5, anchor=11, extract_single_season=True)
        elif self._frequency == "apr-oct":
            tasks = bin_seasonal(cells, months=7, anchor=4, extract_single_season=True)
        elif self._frequency == "annual-fy":
            tasks = bin_seasonal(cells, months=12, anchor=7)
        elif self._frequency == "annual":
            tasks = bin_annual(cells)
        elif temporal_range is not None:
            tasks = bin_generic(cells, [temporal_range])
        else:
            tasks = bin_annual(cells)
        # Remove duplicate source uuids.
        # Duplicates occur when queried datasets are captured around UTC midnight
        # and around weekly boundary
        tasks = {k: set(dss) for k, dss in tasks.items()}
        tasks_uuid = {k: [ds.id for ds in dss] for k, dss in tasks.items()}

        all_ids = set()
        for k, dss in tasks_uuid.items():
            all_ids.update(dss)
        msg(f"Total of {len(all_ids):,d} unique dataset IDs after filtering")

        if len(all_ids) == 0:
            return True

        msg(f"Saving tasks to disk ({len(tasks)})")
        cache.add_grid_tiles(self._grid, tasks_uuid)
        msg(".. done")

        self._write_info(tasks, msg, cells, debug)

        return True

    def _write_info(self, tasks, msg, cells, debug):
        # pylint:disable=too-many-locals
        csv_path = self.out_path(".csv")
        msg(f"Writing summary to {csv_path}")
        with open(csv_path, "wt", encoding="utf8") as f:
            f.write('"T","X","Y","datasets","days"\n')

            for p, x, y in sorted(tasks):
                dss = tasks[(p, x, y)]
                n_dss = len(dss)
                n_days = len(set(ds.time.date() for ds in dss))
                line = f'"{p}", {x:+05d}, {y:+05d}, {n_dss:4d}, {n_days:4d}\n'
                f.write(line)

        msg("Dumping GeoJSON(s)")
        grid_info = compute_grid_info(
            cells, resolution=max(self._gridspec.tile_size) / 4
        )
        tasks_geo = gjson_from_tasks(tasks, grid_info)
        for temporal_range, gjson in tasks_geo.items():
            fname = self.out_path(f"-{temporal_range}.geojson")
            msg(f"..writing to {fname}")
            with open(fname, "wt", encoding="utf8") as f:
                json.dump(gjson, f)

        if debug:
            pkl_path = self.out_path("-cells.pkl")
            msg(f"Saving debug info to: {pkl_path}")
            with open(pkl_path, "wb") as fb:
                pickle.dump(cells, fb)

            pkl_path = self.out_path("-tasks.pkl")
            msg(f"Saving debug info to: {pkl_path}")
            with open(pkl_path, "wb") as fb:
                pickle.dump(tasks, fb)

        return True


class TaskReader:
    def __init__(
        self,
        cache: Union[str, DatasetCache],
        product: Optional[OutputProduct] = None,
        resolution: Optional[Tuple[float, float]] = None,
    ):
        self._cache_path = None

        if len(cache) != 0 and isinstance(cache, str):
            if cache.startswith("s3://"):
                self._cache_path = s3_download(cache)
                cache = self._cache_path
            cache = DatasetCache.open_ro(cache)

            # TODO: verify this things are set in the file
            cfg = cache.get_info_dict("stats/config")
            grid = cfg["grid"]
            gridspec = cache.grids[grid]
        else:  # if read from message, there is no filedb at beginning
            cfg = {}

        self._product = product
        # if not from_sqs, the resolution check can finish before init TaskReader
        self.resolution = resolution
        self._dscache = cache
        self._cfg = cfg
        self._grid = grid if cache else ""
        self._gridspec = gridspec if cache else ""
        self._all_tiles = sorted(idx for idx, _ in cache.tiles(grid)) if cache else []

    def is_compatible_resolution(self, resolution: Tuple[float, float], tol=1e-8):
        for res, sz in zip(resolution, self._gridspec.tile_size):
            res = abs(res)
            npix = int(sz / res)
            if abs(npix * res - sz) > tol:
                return False
        return True

    def change_resolution(self, resolution: Tuple[float, float]):
        """
        Modify GridSpec to have different pixel resolution but still covering same tiles as the original.
        """
        if not self.is_compatible_resolution(resolution):
            raise ValueError(
                "Supplied resolution is not compatible with the current GridSpec"
            )
        gs = self._gridspec
        self._gridspec = GridSpec(
            gs.crs, gs.tile_size, resolution=resolution, origin=gs.origin
        )

    def init_from_sqs(self, local_db_path: str):
        """
        Adding the missing _grid, _gridspec, _gridspec and _all_tiles which skip for sqs task init.
        Upading the cfg which used placeholder filedb path for sqs task init.
        """

        _log = logging.getLogger(__name__)

        cache = DatasetCache.open_ro(local_db_path)

        # TODO: Validate this information. Assumption is the ...
        cfg = cache.get_info_dict("stats/config")
        grid = cfg["grid"]
        gridspec = cache.grids[grid]

        cfg["filedb"] = cache

        # Configure everything on the TaskReader
        self._dscache = cache
        self._cfg = cfg
        self._grid = grid
        self._gridspec = gridspec
        self._all_tiles = sorted(idx for idx, _ in cache.tiles(grid)) if cache else []

        # first time to access the filedb, then it can do the resolution check
        if self.resolution is not None:
            _log.info(
                "Changing resolution to %s, %s", self.resolution[0], self.resolution[1]
            )
            if self.is_compatible_resolution(self.resolution):
                self.change_resolution(self.resolution)
            else:  # if resolution has issue, stop init
                _log.error(
                    "Requested resolution is not compatible with GridSpec in '%s'",
                    cfg.filedb,
                )
                raise ValueError(
                    f"Requested resolution is not compatible with GridSpec in '{cfg.filedb}'"
                )

    def __del__(self):
        if self._cache_path is not None:
            os.unlink(self._cache_path)

    def __repr__(self) -> str:
        grid, path, n = self._grid, str(self._dscache.path), len(self._all_tiles)
        return f"<{path}> grid:{grid} n:{n:,d}"

    def _resolve_product(self, product: Optional[OutputProduct]) -> OutputProduct:
        if product is None:
            product = self._product

        if product is None:
            raise ValueError("Product is not supplied and default is not set")
        return product

    @property
    def cache(self) -> DatasetCache:
        return self._dscache

    @property
    def grid(self) -> GridSpec:
        return self._grid

    @property
    def product(self) -> OutputProduct:
        return self._resolve_product(None)

    @property
    def all_tiles(self) -> List[TileIdx_txy]:
        return self._all_tiles

    def datasets(self, tile_index: TileIdx_txy) -> Tuple[Dataset, ...]:
        return tuple(
            ds for ds in self._dscache.stream_grid_tile(tile_index, self._grid)
        )

    def load_task(
        self,
        tile_index: TileIdx_txy,
        product: Optional[OutputProduct] = None,
        source: Any = None,
        ds_filters: Optional[str] = None,
    ) -> Task:
        product = self._resolve_product(product)

        dss = self.datasets(tile_index)
        if ds_filters is not None:
            ds_checker = DatasetChecker(ds_filters)
            dss = tuple(ds for ds in dss if ds_checker.check_dataset(ds))
        tidx_xy = _xy(tile_index)

        return Task(
            product=product,
            tile_index=tidx_xy,
            geobox=self._gridspec.tile_geobox(tidx_xy),
            time_range=DateTimeRange(tile_index[0]),
            datasets=dss,
            source=source,
        )

    def stream(
        self,
        tiles: Iterable[TileIdx_txy],
        product: Optional[OutputProduct] = None,
        ds_filters: Optional[str] = None,
    ) -> Iterator[Task]:
        product = self._resolve_product(product)
        for tidx in tiles:
            yield self.load_task(tidx, product, ds_filters=ds_filters)

    def stream_from_sqs(
        self,
        sqs_queue,
        product: Optional[OutputProduct] = None,
        visibility_timeout: int = 300,
        ds_filters: Optional[str] = None,
        **kw,
    ) -> Iterator[Task]:
        from odc.aws.queue import get_messages, get_queue
        from ._sqs import SQSWorkToken

        product = self._resolve_product(product)

        if isinstance(sqs_queue, str):
            sqs_queue = get_queue(sqs_queue)

        for msg in get_messages(sqs_queue, visibility_timeout=visibility_timeout, **kw):
            token = SQSWorkToken(msg, visibility_timeout)
            tidx, filedb = parse_sqs(msg.body)
            local_db_file = None

            # Avoid downloading the file multiple times.
            if urlparse(filedb).scheme == "s3":
                _, key = s3_url_parse(filedb)
                local_db_file = key.split("/")[-1]
                self._cache_path = local_db_file

                # Make sure we have this DB downloaded
                if not os.path.isfile(local_db_file):
                    s3_download(filedb, destination=local_db_file)
            else:
                # Assume it's a local file if it's not an S3 URL
                local_db_file = filedb

            # Initialise TaskReader information from the DB
            self.init_from_sqs(local_db_file)

            # Load the task
            yield self.load_task(tidx, product, source=token, ds_filters=ds_filters)


class DatasetChecker:
    def __init__(self, ds_filters):
        ds_filters = ds_filters.split("|")
        self.ds_filters = tuple(json.loads(ds_filter) for ds_filter in ds_filters)

    @staticmethod
    def check_dt(ds_filter, datetime_str):
        time_range = DateTimeRange(ds_filter["datetime"])
        dt = ciso8601.parse_datetime_as_naive(datetime_str)
        return dt in time_range

    def check_ds_1(self, ds_filter, ds):
        valid = True
        for key in ds_filter.keys():
            if key == "datetime":
                valid &= self.check_dt(ds_filter, ds.metadata_doc["properties"][key])
            else:
                valid &= ds_filter[key] == ds.metadata_doc["properties"][key]

        return valid

    def check_dataset(self, ds):
        valid = False
        for ds_filter in self.ds_filters:
            valid |= self.check_ds_1(ds_filter, ds)

        return valid



"""
Water Observations Summaries

Water Observations Summaries are made up of:

- `count_clear`: a count of every time a pixel was observed
                (not obscured by terrain or clouds)
- `count_wet`: a count of every time a pixel was observed and wet
- `frequency`: what fraction of time (wet/clear) was the pixel wet

The counts are stored as `int16` and the frequency as `float32`.

There are two different Stats Plugin classes implemented in this module.
The first generates summary data from individual water observations,
and the second generates a summary of summaries, which is used when generating an all
of time summary from existing annual summaries.

"""
from typing import Dict, Tuple, Iterable
import numpy as np
import xarray as xr
from odc.algo import safe_div, apply_numexpr, keep_good_only
from ._registry import StatsPluginInterface, register
from odc.algo._masking import _or_fuser, mask_cleanup


class StatsWofs(StatsPluginInterface):
    """
    Generate a Summary of Water Observations data from individual observations

    The summary is made up of counts of visible and visible and wet,
    and the frequency of visible and wet.

    Output data types are:
    - `count_clear`: `int16`
    - `count_wet`: `int16`
    - `frequency`: `float32`

    Special care is taken when handling NaN values and `no-data` values.
    """

    NAME = "ga_ls_wo_summary"
    SHORT_NAME = NAME
    VERSION = "1.6.1"
    PRODUCT_FAMILY = "wo_summary"

    # these get padded out if dilation was requested
    BAD_BITS_MASK = dict(
        cloud=(1 << 6), cloud_shadow=(1 << 5), terrain_shadow=(1 << 3)
    )  # Cloud/Shadow + Terrain Shadow

    def __init__(
        self, cloud_filters: Dict[str, Iterable[Tuple[str, int]]] = None, **kwargs
    ):
        super().__init__(input_bands=["water"], **kwargs)
        self.cloud_filters = cloud_filters if cloud_filters is not None else {}

    @property
    def measurements(self) -> Tuple[str, ...]:
        return "count_wet", "count_clear", "frequency"

    def native_transform(self, xx):
        """
        xx.water -- uint8 classifier bitmask

        .. code-block::

          |128| 64| 32| 16| |  8|  4|  2|  1|
          |---|---|---|---|=|---|---|---|---|
            7   6   5   4     3   2   1   0
            |   |   |   |     |   |   |   |
            |   |   |   |     |   |   |   x---> NODATA: 1 -- all bands were nodata
            |   |   |   |     |   |   o-------> Non Contiguous - some bands were nodata)
            |   |   |   |     |   x-----------> Low Solar Angle
            |   |   |   |     o---------------> Terrain Shadow
            |   |   |   |
            |   |   |   x---------------------> Terrain High Slope
            |   |   o-------------------------> Cloud Shadow
            |   x-----------------------------> Cloud
            o---------------------------------> Water

        out:
          .bad<Bool>   - non-clear pixel should not be counted
          .some<Bool>  - there is data (x.water & 0b1) == 0,
                         to distinguish "count=0" resulted from "nodata"
                         or "non-clear" = bad + dry + wet
          .dry<Bool>   - pixel has dry classification and is not ``bad``
          .wet<Bool>   - pixel has wet classification and is not ``bad``
        """
        xx["bad"] = (xx.water & 1) == 0
        xx["bad"] &= (xx.water & (~(1 << 7) | 1)) > 0  # bad

        # dilate 'bad'
        for key, val in self.BAD_BITS_MASK.items():
            if self.cloud_filters.get(key) is not None:
                raw_mask = (xx["water"] & val) > 0
                raw_mask = mask_cleanup(
                    raw_mask, mask_filters=self.cloud_filters.get(key)
                )
                xx["bad"] |= raw_mask

        xx["dry"] = (xx.water == 0) & ~xx["bad"]
        xx["wet"] = (xx.water == 128) & ~xx["bad"]
        xx = xx.drop_vars("water")
        for dv in xx.data_vars.values():
            dv.attrs.pop("nodata", None)

        return xx

    def fuser(self, xx):
        return _or_fuser(xx)

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        """
        bad + anything -> bad
        wet + wet/nodata -> wet
        dry + dry/nodata -> dry
        nodata + anything -> anything
        """

        nodata = -999

        wet = apply_numexpr("wet & (~dry) & (~bad)", xx, dtype="bool")
        dry = apply_numexpr("dry & (~wet) & (~bad)", xx, dtype="bool")

        count_wet = wet.sum(axis=0, dtype="int16")
        count_dry = dry.sum(axis=0, dtype="int16")
        count_clear = count_wet + count_dry
        frequency = safe_div(count_wet, count_clear, dtype="float32")
        count_some = xx.bad.sum(axis=0, dtype="int16") + count_clear

        count_wet.attrs["nodata"] = nodata
        count_clear.attrs["nodata"] = nodata

        is_ok = count_some > 0
        count_wet = keep_good_only(count_wet, is_ok)
        count_clear = keep_good_only(count_clear, is_ok)

        return xr.Dataset(
            dict(
                count_wet=count_wet,
                count_clear=count_clear,
                frequency=frequency,
            )
        )


register("wofs-summary", StatsWofs)


class StatsWofsFullHistory(StatsPluginInterface):
    """
    Generate a Summary of Water Observations data from existing WO Summaries

    This is useful to turn Annual summary data into all of time summaries.

    Output data is the same as `StatsWofs` produces.
    - `count_clear`: `int16`
    - `count_wet`: `int16`
    - `frequency`: `float32`

    Special care is taken with no-data values, both to pass them through
    and when calculating the counts and frequencies.
    """

    NAME = "ga_ls_wo_fq_myear_3"
    SHORT_NAME = NAME
    VERSION = "1.6.0"
    PRODUCT_FAMILY = "wo_summary"

    def __init__(self, **kwargs):
        super().__init__(input_bands=["count_wet", "count_clear"])

    @property
    def measurements(self) -> Tuple[str, ...]:
        return "count_wet", "count_clear", "frequency"

    def fuser(self, xx):
        """
        no fuse required since group by none
        return loaded data
        """
        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        dtype = xx.count_clear.dtype
        nodata = dtype.type(xx.count_clear.nodata)

        # `missing` is a record of all pixels that were never observed.
        # Store it separately first, then substitute it back in
        # after computing the counts and  frequency.
        missing = (xx.count_clear == xx.count_clear.nodata).all(axis=0)
        cc = apply_numexpr(
            "where(count_clear==nodata, 0, count_clear)",
            xx,
            dtype="int16",
            casting="unsafe",
            nodata=nodata,
        ).sum(axis=0, dtype=dtype)

        cw = apply_numexpr(
            "where(count_wet==nodata, 0, count_wet)",
            xx,
            dtype=dtype,
            casting="unsafe",
            nodata=nodata,
        ).sum(axis=0, dtype=dtype)

        _yy = xr.Dataset(dict(cc=cc, cw=cw, missing=missing))

        frequency = apply_numexpr(
            "where(cc==0, _nan, (_1*cw)/(_1*cc))",
            _yy,
            dtype="float32",
            _1=np.float32(1),
            _nan=np.float32("nan"),
        )

        # Finalise the *count* variables by re-inserting the no-data value
        # based on `missing`
        count_clear = apply_numexpr(
            "where(missing, nodata, cc)",
            _yy,
            dtype=dtype,
            nodata=nodata,
            casting="unsafe",
        )
        count_wet = apply_numexpr(
            "where(missing, nodata, cw)",
            _yy,
            dtype=dtype,
            nodata=nodata,
            casting="unsafe",
        )

        count_clear.attrs["nodata"] = int(nodata)
        count_wet.attrs["nodata"] = int(nodata)

        yy = xr.Dataset(
            dict(count_clear=count_clear, count_wet=count_wet, frequency=frequency)
        )
        return yy


register("wofs-summary-fh", StatsWofsFullHistory)



"""
Geomedian
"""
from typing import Optional, Union, Tuple, Iterable, Dict
import xarray as xr
from odc.algo import geomedian_with_mads
from ._registry import StatsPluginInterface, register
from odc.algo import enum_to_bool, erase_bad
from odc.algo import mask_cleanup
import logging

_log = logging.getLogger(__name__)


class StatsGM(StatsPluginInterface):
    NAME = "gm"
    SHORT_NAME = NAME
    VERSION = "0.0.0"
    PRODUCT_FAMILY = "geomedian"

    def __init__(
        self,
        bands: Tuple[str, ...],
        mask_band: str,
        contiguity_band: Optional[str] = None,
        nodata_classes: Optional[Tuple[str, ...]] = None,
        cloud_filters: Dict[
            Union[str, Tuple[str, ...]], Iterable[Tuple[str, int]]
        ] = None,
        basis_band=None,
        aux_names: Dict[str, str] = None,
        work_chunks: Tuple[int, int] = (400, 400),
        **kwargs,
    ):
        aux_names = (
            dict(smad="smad", emad="emad", bcmad="bcmad", count="count")
            if aux_names is None
            else aux_names
        )
        self.bands = tuple(bands)
        self._mask_band = mask_band
        self._contiguity_band = contiguity_band
        if nodata_classes is not None:
            nodata_classes = tuple(nodata_classes)
        self._nodata_classes = nodata_classes
        input_bands = self.bands
        if self._nodata_classes is not None:
            # NOTE: this ends up loading Mask band twice, once to compute
            # ``.erase`` band and once to compute ``nodata`` mask.
            input_bands = (*input_bands, self._mask_band)

        super().__init__(
            input_bands=input_bands,
            basis=basis_band or self.bands[0],
            **kwargs,
        )

        self.cloud_filters = cloud_filters
        self._renames = aux_names
        self.aux_bands = tuple(
            self._renames.get(k, k)
            for k in (
                "smad",
                "emad",
                "bcmad",
                "count",
            )
        )

        self._work_chunks = work_chunks

    @property
    def measurements(self) -> Tuple[str, ...]:
        return self.bands + self.aux_bands

    def native_transform(self, xx: xr.Dataset) -> xr.Dataset:
        if self._mask_band not in xx.data_vars:
            return xx

        # Erase Data Pixels for which mask == nodata
        mask = xx[self._mask_band]
        bad = enum_to_bool(mask, self._nodata_classes)
        # Apply the contiguity flag
        if self._contiguity_band is not None:
            non_contiguent = xx.get(self._contiguity_band, 1) == 0
            bad = bad | non_contiguent

        if self.cloud_filters is not None:
            for cloud_class, c_filter in self.cloud_filters.items():
                if not isinstance(cloud_class, tuple):
                    cloud_class = (cloud_class,)
                cloud_mask = enum_to_bool(mask, cloud_class)
                cloud_mask_buffered = mask_cleanup(cloud_mask, mask_filters=c_filter)
                bad = cloud_mask_buffered | bad
        else:
            cloud_shadow_mask = enum_to_bool(mask, ("cloud", "shadow"))
            bad = cloud_shadow_mask | bad
            _log.info("Applying cloud/shadow mask without buffering.")

        if self._contiguity_band is not None:
            xx = xx.drop_vars([self._mask_band] + [self._contiguity_band])
        else:
            xx = xx.drop_vars([self._mask_band])
        xx = erase_bad(xx, bad)
        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        scale = 1 / 10_000
        cfg = dict(
            maxiters=1000,
            num_threads=1,
            scale=scale,
            offset=-1 * scale,
            reshape_strategy="mem",
            out_chunks=(-1, -1, -1),
            work_chunks=self._work_chunks,
            compute_count=True,
            compute_mads=True,
        )

        gm = geomedian_with_mads(xx, **cfg)
        gm = gm.rename(self._renames)

        return gm


register("gm-generic", StatsGM)


class StatsGMS2(StatsGM):
    NAME = "gm_s2_annual"
    SHORT_NAME = NAME
    VERSION = "0.0.0"
    PRODUCT_FAMILY = "geomedian"
    DEFAULT_FILTER = [("opening", 2), ("dilation", 5)]

    def __init__(
        self,
        bands: Optional[Tuple[str, ...]] = None,
        mask_band: str = "SCL",
        nodata_classes: Optional[Tuple[str, ...]] = ("no data",),
        cloud_filters: Dict[
            Union[str, Tuple[str, ...]], Iterable[Tuple[str, int]]
        ] = None,
        aux_names: Dict[str, str] = None,
        rgb_bands=None,
        **kwargs,
    ):
        cloud_filters = (
            {
                (
                    "cloud shadows",
                    "cloud medium probability",
                    "cloud high probability",
                    "thin cirrus",
                ): self.DEFAULT_FILTER,
            }
            if cloud_filters is None
            else cloud_filters
        )

        aux_names = (
            dict(smad="SMAD", emad="EMAD", bcmad="BCMAD", count="COUNT")
            if aux_names is None
            else aux_names
        )

        if bands is None:
            bands = (
                "B02",
                "B03",
                "B04",
                "B05",
                "B06",
                "B07",
                "B08",
                "B8A",
                "B11",
                "B12",
            )
            if rgb_bands is None:
                rgb_bands = ("B04", "B03", "B02")

        super().__init__(
            bands=bands,
            mask_band=mask_band,
            nodata_classes=nodata_classes,
            cloud_filters=cloud_filters,
            aux_names=aux_names,
            rgb_bands=rgb_bands,
            **kwargs,
        )


register("gm-s2", StatsGMS2)


class StatsGMLS(StatsGM):
    NAME = "gm_ls_annual"
    SHORT_NAME = NAME
    VERSION = "3.0.1"
    PRODUCT_FAMILY = "geomedian"

    def __init__(
        self,
        bands: Optional[Tuple[str, ...]] = None,
        mask_band: str = "fmask",
        contiguity_band: str = "nbart_contiguity",
        nodata_classes: Optional[Tuple[str, ...]] = ("nodata",),
        cloud_filters: Dict[
            Union[str, Tuple[str, ...]], Iterable[Tuple[str, int]]
        ] = None,
        aux_names: Dict[str, str] = None,
        rgb_bands=None,
        **kwargs,
    ):
        aux_names = (
            dict(
                smad="sdev",
                emad="edev",
                bcmad="bcdev",
                count="count",
            )
            if aux_names is None
            else aux_names
        )

        if bands is None:
            bands = (
                "nbart_red",
                "nbart_green",
                "nbart_blue",
                "nbart_nir",
                "nbart_swir_1",
                "nbart_swir_2",
                "nbart_contiguity",
            )
            if rgb_bands is None:
                rgb_bands = ("nbart_red", "nbart_green", "nbart_blue")

        super().__init__(
            bands=bands,
            mask_band=mask_band,
            contiguity_band=contiguity_band,
            cloud_filters=cloud_filters,
            nodata_classes=nodata_classes,
            aux_names=aux_names,
            rgb_bands=rgb_bands,
            **kwargs,
        )

    @property
    def measurements(self) -> Tuple[str, ...]:
        return (
            tuple(b for b in self.bands if b != self._contiguity_band) + self.aux_bands
        )


register("gm-ls", StatsGMLS)



import json

import click
import sys
import logging
from datacube import Datacube
from .tasks import SaveTasks
from .model import DateTimeRange

from ._cli_common import main, click_range2d, click_yaml_cfg, setup_logging

CONFIG_ITEMS = [
    "grid",
    "frequency",
    "complevel",
    "overwrite",
    "gqa",
    "input_products",
    "dataset_filter",
]


@main.command("save-tasks")
@click.option(
    "--grid",
    type=str,
    help=(
        "Grid name or spec: au-{10|20|30|60},africa-{10|20|30|60},"
        "albers-au-25 (legacy one) 'crs;pixel_resolution;shape_in_pixels'"
    ),
    default=None,
)
@click.option(
    "--year",
    type=int,
    help=(
        "Only extract datasets for a given year."
        "This is a shortcut for --temporal-range=<int>--P1Y"
    ),
)
@click.option(
    "--temporal-range",
    type=str,
    help=(
        "Only extract datasets for a given time range,"
        "Example '2020-05--P1M' month of May 2020"
    ),
)
@click.option(
    "--frequency",
    type=str,
    help=(
        "Specify temporal binning: "
        "annual|annual-fy|semiannual|seasonal|quartely|3month-seasons|rolling-3months|nov-mar|apr-oct|all"
    ),
)
@click.option("--env", "-E", type=str, help="Datacube environment name")
@click.option(
    "-z",
    "complevel",
    type=int,
    default=6,
    help="Compression setting for zstandard 1-fast, 9+ good but slow",
)
@click.option(
    "--overwrite", is_flag=True, default=None, help="Overwrite output if it exists"
)
@click.option(
    "--tiles", help='Limit query to tiles example: "0:3,2:4"', callback=click_range2d
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    hidden=True,
    help="Dump debug data to pickle",
)
@click.option(
    "--gqa",
    type=float,
    help="Only save datasets that pass `gqa_iterative_mean_xy <= gqa` test",
)
@click.option(
    "--usgs-collection-category",
    type=str,
    help=(
        "Only save datasets that pass "
        "`collection_category == usgs_collection_category` test"
    ),
)
@click.option(
    "--dataset-filter",
    type=str,
    default=None,
    help='Filter to apply on datasets - {"collection_category": "T1"}',
)
@click_yaml_cfg("--config", help="Save tasks Config")
@click.option("--input-products", type=str, default="")
@click.argument("output", type=str, nargs=1, default="")
# pylint: disable=too-many-arguments, too-many-locals
# pylint: disable=too-many-branches, too-many-statements
def save_tasks(
    config,
    grid,
    year,
    temporal_range,
    frequency,
    output,
    input_products,
    dataset_filter,
    env,
    complevel,
    overwrite,
    tiles=None,
    debug=False,
    gqa=None,
    usgs_collection_category=None,
):
    """
    Prepare tasks for processing (query db).

    <todo more help goes here>

    \b
    Not yet implemented features:
      - output

    """
    setup_logging()

    _log = logging.getLogger(__name__)
    if temporal_range is not None and year is not None:
        print("Can only supply one of --year or --temporal_range", file=sys.stderr)
        sys.exit(1)

    if config is None:
        config = {}

    _cfg = {k: config.get(k) for k in CONFIG_ITEMS if config.get(k) is not None}

    cfg_from_cli = {
        k: v
        for k, v in dict(
            grid=grid,
            frequency=frequency,
            gqa=gqa,
            input_products=input_products,
            complevel=complevel,
            dataset_filter=dataset_filter,
            overwrite=overwrite,
        ).items()
        if v is not None and v != ""
    }

    _log.info("Config overrides: %s", cfg_from_cli)
    _cfg.update(cfg_from_cli)
    _log.info("Using config: %s", _cfg)

    gqa = _cfg.pop("gqa", None)
    input_products = _cfg.pop("input_products", None)
    dataset_filter = _cfg.pop("dataset_filter", None)

    if input_products is None:
        print("Input products has to be specified", file=sys.stderr)
        sys.exit(1)

    if _cfg.get("grid") is None:
        print(
            "grid must  be  one of au-{10|20|30|60}, africa-{10|20|30|60}, \
             albers_au_25 (legacy one) or custom like 'epsg:3857;30;5000' \
             (30m pixels 5,000 per side in epsg:3857) ",
            file=sys.stderr,
        )
        sys.exit(1)

    if _cfg.get("frequency") is not None:
        if _cfg.get("frequency") not in (
            "annual",
            "annual-fy",
            "semiannual",
            "seasonal",
            "quartely",
            "3month-seasons",
            "rolling-3months",
            "nov-mar",
            "apr-oct",
            "all",
        ):
            print(
                f"""Frequency must be one of annual|annual-fy|semiannual|seasonal|
                quartely|3month-seasons|rolling-3months|nov-mar|apr-oct|all
                and not '{frequency}'""",
                file=sys.stderr,
            )
            sys.exit(1)

    if temporal_range is not None:
        try:
            temporal_range = DateTimeRange(temporal_range)
        except ValueError:
            print(
                f"Failed to parse supplied temporal_range: '{temporal_range}'",
                file=sys.stderr,
            )
            sys.exit(1)

    if year is not None:
        temporal_range = DateTimeRange.year(year)

    if output == "":
        if temporal_range is not None:
            output = f"{input_products}_{temporal_range.short}.db"
        else:
            output = f"{input_products}_all.db"

    try:
        tasks = SaveTasks(output, **_cfg)
    except ValueError as e:
        print(str(e), file=sys.stderr)
        sys.exit(1)

    def on_message(msg):
        print(msg)

    def gqa_predicate(ds):
        return ds.metadata.gqa_iterative_mean_xy <= gqa

    def collection_category_predicate(ds):
        if ds.type.name in ["ls5_sr", "ls7_sr", "ls8_sr", "ls9_sr"]:
            return ds.metadata.collection_category == usgs_collection_category
        else:
            return True

    predicate = None
    # These two are exclusive. GQA is from DEA, whereas collection_category is from USGS
    if gqa is not None:
        predicate = gqa_predicate
    if usgs_collection_category is not None:
        predicate = collection_category_predicate

    ds_filter = {}
    if dataset_filter:
        ds_filter = json.loads(dataset_filter)

    dc = Datacube(env=env)
    try:
        ok = tasks.save(
            dc,
            input_products,
            dataset_filter=ds_filter,
            temporal_range=temporal_range,
            tiles=tiles,
            predicate=predicate,
            debug=debug,
            msg=on_message,
        )
    except ValueError as e:
        print(str(e))
        sys.exit(2)

    if not ok:
        # exit with error code, failure message was already printed
        sys.exit(3)



"""
Fractional Cover Percentiles
"""
from functools import partial
from itertools import product
from typing import Tuple, Dict, Iterable, Optional

import numpy as np
import xarray as xr
from odc.algo import keep_good_only
from odc.algo._masking import _fuse_mean_np, _fuse_or_np, _xr_fuse, mask_cleanup
from odc.algo._percentile import xr_quantile_bands

from ._registry import StatsPluginInterface, register

NODATA = 255


class StatsFCP(StatsPluginInterface):
    NAME = "ga_fc_percentiles"
    SHORT_NAME = NAME
    VERSION = "0.0.3"
    PRODUCT_FAMILY = "fc_percentiles"

    BAD_BITS_MASK = dict(cloud=(1 << 6), cloud_shadow=(1 << 5), terrain_shadow=(1 << 3))

    def __init__(
        self,
        max_sum_limit: Optional[int] = None,
        clip_range: Optional[Tuple] = None,
        ue_threshold: Optional[int] = None,
        count_valid: Optional[bool] = False,
        cloud_filters: Dict[str, Iterable[Tuple[str, int]]] = None,
        **kwargs,
    ):
        super().__init__(input_bands=["water", "pv", "bs", "npv", "ue"], **kwargs)

        self.max_sum_limit = max_sum_limit
        self.clip_range = clip_range
        self.ue_threshold = ue_threshold
        self.count_valid = count_valid
        self.cloud_filters = cloud_filters if cloud_filters is not None else {}

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurements = [
            f"{b}_pc_{p}" for b, p in product(["pv", "bs", "npv"], ["10", "50", "90"])
        ]
        _measurements.append("qa")
        if self.count_valid:
            _measurements.append("count_valid")
        return _measurements

    def native_transform(self, xx):
        """
        Loads data in its native projection. It performs the following:

        1. Load all fc and WOfS bands
        2. Set the high terrain slope flag to 0
        3. Set all pixels that are not clear and dry to NODATA
        4. Calculate the clear wet pixels
        5. Drop the WOfS band
        """

        # not mask against bit 4: terrain high slope
        # Pick out the dry and wet pixels
        valid = (xx["water"] & ~(1 << 4)) == 0
        wet = (xx["water"] & ~(1 << 4)) == 128

        # dilate both 'valid' and 'water'
        for key, val in self.BAD_BITS_MASK.items():
            if self.cloud_filters.get(key) is not None:
                raw_mask = (xx["water"] & val) > 0
                raw_mask = mask_cleanup(
                    raw_mask, mask_filters=self.cloud_filters.get(key)
                )
                valid &= ~raw_mask
                wet &= ~raw_mask

        xx = xx.drop_vars(["water"])

        # Pick out the pixels that have an unmixing error of less than the threshold
        if self.ue_threshold is not None:
            # No QA
            valid &= xx.ue < self.ue_threshold
        xx = xx.drop_vars(["ue"])

        # If there's a sum limit or clip range, implement these
        if self.max_sum_limit is not None or self.clip_range is not None:
            sum_bands = 0
            for band in xx.data_vars.keys():
                attributes = xx[band].attrs
                mask = xx[band] == NODATA
                band_data = keep_good_only(xx[band], ~mask, nodata=0)

                if self.max_sum_limit is not None:
                    sum_bands = sum_bands + band_data

                if self.clip_range is not None:
                    # No QA
                    clipped = np.clip(xx[band], self.clip_range[0], self.clip_range[1])
                    # Set masked values back to NODATA
                    xx[band] = clipped.where(~mask, NODATA)
                    xx[band].attrs = attributes

            if self.max_sum_limit is not None:
                valid &= sum_bands < self.max_sum_limit

        xx = keep_good_only(xx, valid, nodata=NODATA)
        xx["wet"] = wet
        xx["valid"] = valid

        return xx

    def fuser(self, xx):
        wet = xx["wet"]
        valid = xx["valid"]

        xx = _xr_fuse(
            xx.drop_vars(["wet", "valid"]), partial(_fuse_mean_np, nodata=NODATA), ""
        )

        xx["wet"] = _xr_fuse(wet, _fuse_or_np, wet.name)
        xx["valid"] = _xr_fuse(valid, _fuse_or_np, valid.name)

        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        # (!all_bands_valid) & is_ever_wet => 0
        # (!all_bands_valid) & (!is_ever_wet) => 1
        # all_bands_valid => 2

        wet = xx["wet"]
        valid = xx["valid"]
        xx = xx.drop_vars(["wet", "valid"])

        yy = xr_quantile_bands(xx, [0.1, 0.5, 0.9], nodata=NODATA)
        # wet and valid(dry) are mutually exclusive, here we keep good data
        wet &= ~valid
        is_ever_wet = (wet.sum(axis=0, dtype="int16") > 0).astype("uint8")

        all_bands_valid = True
        for band in yy.data_vars.keys():
            all_bands_valid &= yy[band] != NODATA

        all_bands_valid = all_bands_valid.astype("uint8")

        yy["qa"] = 1 + all_bands_valid - is_ever_wet * (1 - all_bands_valid)

        if self.count_valid:
            yy["count_valid"] = valid.sum(axis=0, dtype="int16")

        return yy


register("fc-percentiles", StatsFCP)
